__version__ = "0.1.1"
__author__ = "Shipei Xing"
__email__ = "philipxsp@hotmail.com"
__author_email__ = "philipxsp@hotmail.com"

__copyright__ = "Copyright 2024 Shipei Xing"
__license_expression__ = "Apache 2.0"

__url__ = "https://github.com/Philipbear/ms1_id"

__title__ = "ms1_id"